@@ -1,6 +1,6 @@
 """ PENQUINS - Processing ENormous Queries of kowalski Users INStantaneously """
 
-__all__ = ['Kowalski', '__version__']
+__all__ = ["Kowalski", "__version__"]
 
 
 from copy import deepcopy
@@ -15,20 +15,18 @@ import secrets
 import string
 import traceback
 from tqdm.auto import tqdm
-from typing import Union
+from typing import Mapping, Optional, Sequence, Union
 
 
-__version__ = '2.0.1'
+__version__ = "2.2.0"
 
 
 Num = Union[int, float]
-QueryPart = Union['task', 'result']
-Method = Union['get', 'post', 'put', 'patch', 'delete']
 
 
-DEFAULT_TIMEOUT = 5  # seconds
-DEFAULT_RETRIES = 3
-DEFAULT_BACKOFF_FACTOR = 1
+DEFAULT_TIMEOUT: int = 5  # seconds
+DEFAULT_RETRIES: int = 3
+DEFAULT_BACKOFF_FACTOR: int = 1
 
 
 class TimeoutHTTPAdapter(HTTPAdapter):
@@ -40,36 +38,44 @@ class TimeoutHTTPAdapter(HTTPAdapter):
         super().__init__(*args, **kwargs)
 
     def send(self, request, **kwargs):
-        timeout = kwargs.get("timeout")
-        if timeout is None:
-            kwargs["timeout"] = self.timeout
-        return super().send(request, **kwargs)
+        try:
+            timeout = kwargs.get("timeout")
+            if timeout is None:
+                kwargs["timeout"] = self.timeout
+            return super().send(request, **kwargs)
+        except AttributeError:
+            kwargs["timeout"] = DEFAULT_TIMEOUT
 
 
 class Kowalski:
-    """
-        Class to communicate with a Kowalski instance
-    """
+    """Class to communicate with a Kowalski instance"""
 
     def __init__(
-            self, username=None, password=None, token=None,
-            protocol: str = 'https', host: str = 'kowalski.caltech.edu', port: int = 443,
-            verbose: bool = False,
-            **kwargs
+        self,
+        username: Optional[str] = None,
+        password: Optional[str] = None,
+        token: Optional[str] = None,
+        protocol: str = "https",
+        host: str = "kowalski.caltech.edu",
+        port: int = 443,
+        verbose: bool = False,
+        **kwargs,
     ):
         """
-            username, password, token, protocol, host, port:
-                Kowalski instance access credentials and address
-                If password is omitted, then look up default credentials from
-                the ~/.netrc file.
-            timeout, pool_connections, pool_maxsize, max_retries, backoff_factor, pool_block:
-                control requests.Session connection pool settings
-            verbose:
-                "Status, Kowalski!"
+        username, password, token, protocol, host, port:
+            Kowalski instance access credentials and address
+            If password is omitted, then look up default credentials from
+            the ~/.netrc file.
+        timeout, pool_connections, pool_maxsize, max_retries, backoff_factor, pool_block:
+            control requests.Session connection pool settings
+        verbose:
+            "Status, Kowalski!"
         """
 
         if (username is None) and (password is None) and (token is None):
-            raise ValueError("Missing credentials: provide either username and password or token")
+            raise ValueError(
+                "Missing credentials: provide either username and password or token"
+            )
 
         if (username is not None) and (password is None):
             netrc_auth = netrc().authenticators(host)
@@ -84,7 +90,7 @@ class Kowalski:
         self.host = host
         self.port = port
 
-        self.base_url = f'{self.protocol}://{self.host}:{self.port}'
+        self.base_url = f"{self.protocol}://{self.host}:{self.port}"
 
         self.session = requests.Session()
 
@@ -105,14 +111,14 @@ class Kowalski:
             total=max_retries,
             backoff_factor=backoff_factor,
             status_forcelist=[405, 429, 500, 502, 503, 504],
-            method_whitelist=["HEAD", "GET", "PUT", "POST", "PATCH"]
+            allowed_methods=["HEAD", "GET", "PUT", "POST", "PATCH"],
         )
         adapter = TimeoutHTTPAdapter(
             timeout=timeout,
             max_retries=retries,
             pool_connections=pool_connections,
             pool_maxsize=pool_maxsize,
-            pool_block=pool_block
+            pool_block=pool_block,
         )
         self.session.mount("https://", adapter)
         self.session.mount("http://", adapter)
@@ -125,14 +131,14 @@ class Kowalski:
         else:
             self.token = token
 
-        self.headers = {'Authorization': self.token}
+        self.headers = {"Authorization": self.token}
 
         self.methods = {
-            'get': self.session.get,
-            'post': self.session.post,
-            'put': self.session.put,
-            'patch': self.session.patch,
-            'delete': self.session.delete
+            "get": self.session.get,
+            "post": self.session.post,
+            "put": self.session.put,
+            "patch": self.session.patch,
+            "delete": self.session.delete,
         }
 
     def __enter__(self):
@@ -162,53 +168,64 @@ class Kowalski:
 
         # post username and password, get access token
         auth = self.session.post(
-            f'{self.base_url}/api/auth',
-            json={"username": self.username, "password": self.password, "penquins.__version__": __version__}
+            f"{self.base_url}/api/auth",
+            json={
+                "username": self.username,
+                "password": self.password,
+                "penquins.__version__": __version__,
+            },
         )
 
         if auth.status_code == requests.codes.ok:
             if self.v:
                 print(auth.json())
 
-            if 'token' not in auth.json():
-                print('Authentication failed')
-                raise Exception(auth.json().get('message', 'Authentication failed'))
+            if "token" not in auth.json():
+                print("Authentication failed")
+                raise Exception(auth.json().get("message", "Authentication failed"))
 
-            access_token = auth.json().get('token')
+            access_token = auth.json().get("token")
 
             if self.v:
-                print('Successfully authenticated')
+                print("Successfully authenticated")
 
             return access_token
 
-        raise Exception('Authentication failed')
+        raise Exception("Authentication failed")
 
-    def api(self, data: dict, endpoint: str = None, method: Method = None):
+    def api(self, method: str, endpoint: str, data: Optional[Mapping] = None):
+        """Call API endpoint on Kowalski"""
+        method = method.lower()
+        # allow using both "/<path>/<endpoint>" and "<path>/<endpoint>"
+        endpoint = endpoint[1:] if endpoint.startswith("/") else endpoint
 
         if endpoint is None:
-            raise ValueError('Endpoint not specified')
-        if method not in ['get', 'post', 'put', 'patch', 'delete']:
-            raise ValueError(f'Unsupported method: {method}')
+            raise ValueError("Endpoint not specified")
+        if method not in ["get", "post", "put", "patch", "delete"]:
+            raise ValueError(f"Unsupported method: {method}")
 
-        if method.lower() != 'get':
-            resp = self.methods[method.lower()](
+        if method != "get":
+            resp = self.methods[method](
                 os.path.join(self.base_url, endpoint),
                 json=data,
                 headers=self.headers,
             )
         else:
-            resp = self.methods[method.lower()](
+            resp = self.methods[method](
                 os.path.join(self.base_url, endpoint),
                 params=data,
                 headers=self.headers,
             )
 
-        if resp.status_code == requests.codes.ok:
-            return loads(resp.text)
+        return loads(resp.text)
 
-        raise Exception('API call failed')
+    def batch_query(self, queries: Sequence[Mapping], n_treads: int = 4):
+        """Call Kowalski's /api/queries endpoint using multiple processes
 
-    def batch_query(self, queries, n_treads: int = 4):
+        :param queries: sequence of queries
+        :param n_treads: number of processes to use
+        :return:
+        """
         n_treads = min(len(queries), n_treads)
 
         with ThreadPool(processes=n_treads) as pool:
@@ -216,39 +233,46 @@ class Kowalski:
                 return list(tqdm(pool.imap(self.query, queries), total=len(queries)))
             return list(pool.imap(self.query, queries))
 
-    def query(self, query):
+    def query(self, query: Mapping):
+        """Call Kowalski's /api/queries endpoint using multiple processes
 
+        :param query: query mapping
+        :return:
+        """
         _query = deepcopy(query)
 
         # by default, all queries are not registered in the db and the task/results are stored on disk as json files
         # giving a significant execution speed up. this behaviour can be overridden.
-        save = _query.get('kwargs', dict()).get('save', False)
+        save = _query.get("kwargs", dict()).get("save", False)
 
         if save:
-            if 'kwargs' not in _query:
-                _query['kwargs'] = dict()
-            if '_id' not in _query['kwargs']:
+            if "kwargs" not in _query:
+                _query["kwargs"] = dict()
+            if "_id" not in _query["kwargs"]:
                 # generate a unique hash id and store it in query if saving query in db on Kowalski is requested
-                _id = ''.join(secrets.choice(string.ascii_lowercase + string.digits) for _ in range(32))
+                _id = "".join(
+                    secrets.choice(string.ascii_lowercase + string.digits)
+                    for _ in range(32)
+                )
 
-                _query['kwargs']['_id'] = _id
+                _query["kwargs"]["_id"] = _id
 
         resp = self.session.post(
-            os.path.join(self.base_url, 'api/queries'),
+            os.path.join(self.base_url, "api/queries"),
             json=_query,
-            headers=self.headers
+            headers=self.headers,
         )
 
         return loads(resp.text)
 
     def ping(self) -> bool:
-        """
-            Ping Kowalski
-        :return: True if connection ok, False otherwise
+        """Ping Kowalski
+
+        :return: True if connection is ok, False otherwise
         """
         try:
             resp = self.session.get(
-                os.path.join(self.base_url, ''),
+                os.path.join(self.base_url, ""),
                 headers=self.headers,
             )
 
